''' Easy method to import all the state machine items '''

from martepy.marte2.objects.statemachine.event import MARTe2StateMachineEvent
from martepy.marte2.objects.statemachine.machine import MARTe2StateMachine

__all__ = [
        'MARTe2StateMachineEvent',
        'MARTe2StateMachine'
    ]
